#include "functions.h"
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

typedef struct { uint64_t state;  uint64_t inc; } pcg32_random_t;

uint32_t pcg32_random_r(pcg32_random_t* rng)
{
    uint64_t oldstate = rng->state;
    // Advance internal state
    rng->state = oldstate * 6364136223846793005ULL + (rng->inc|1);
    // Calculate output function (XSH RR), uses old state for max ILP
    uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
    uint32_t rot = oldstate >> 59u;
    return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
}

void loadFunction(int8_t *a, int size)
{
    // use deterministic random numbers for easier debugging
    pcg32_random_t rng = { 0xDEADBEEFULL, 13 };
    for(size_t i = 0; i < size; ++i)
        a[i] = pcg32_random_r(&rng);
}

static inline uint64_t get_cycles()
{
  uint64_t t;
  __asm volatile ("rdtsc" : "=A"(t));
  return t;
}

int main (int argc, char const *argv[])
{
    int size = 1 << 10;
    
    int8_t *result1, *result2, *a1, *a2;
    
    posix_memalign((void **)&a1, 16, size);
    posix_memalign((void **)&a2, 16, size);
    
    posix_memalign((void **)&result1, 16, size);
    posix_memalign((void **)&result2, 16, size);
    
    loadFunction(a1, size);
    loadFunction(a2, size);
    
    for(size_t i = 0; i < 10; ++i) {
        uint64_t t1 = get_cycles();
        calculateFunction(a1, a2, result1, size);
        uint64_t t2 = get_cycles();
        sseFunction(a1, a2, result2, size);
        uint64_t t3 = get_cycles();
        
        printf("%d %d ", (int)(t2 - t1), (int)(t3 - t2));

        if(memcmp(result1, result2, size)) {
            printf("INCORRECT\n");
            for (size_t j = 0; j < size; ++j) {
                if (result1[j] != result2[j]) {
                    printf("At pos %d, found %d should be %d\n", (int)j, result1[j], result2[j]);
                    break;
                }
            }
        }
        else {
            printf("CORRECT\n");
        }
    }
    return 0;
}
