#include "functions.h"
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

typedef struct { uint64_t state;  uint64_t inc; } pcg32_random_t;

uint32_t pcg32_random_r(pcg32_random_t* rng)
{
    uint64_t oldstate = rng->state;
    // Advance internal state
    rng->state = oldstate * 6364136223846793005ULL + (rng->inc|1);
    // Calculate output function (XSH RR), uses old state for max ILP
    uint32_t xorshifted = ((oldstate >> 18u) ^ oldstate) >> 27u;
    uint32_t rot = oldstate >> 59u;
    return (xorshifted >> rot) | (xorshifted << ((-rot) & 31));
}

void loadFunction(char *a, int size)
{
    // use deterministic random numbers for easier debugging
    pcg32_random_t rng = { 0xDEADBEEFULL, 13 };
    for(size_t i = 0; i < size; ++i)
        a[i] = pcg32_random_r(&rng);
}

static inline uint64_t get_cycles()
{
  uint64_t t;
  __asm volatile ("rdtsc" : "=A"(t));
  return t;
}

int main (int argc, char const *argv[])
{
    int size = 1 << 14;
    
    char *a1;
    
    posix_memalign((void **)&a1, 16, size);
    
    loadFunction(a1, size);
    
    for(size_t i = 0; i < 10; ++i) {
        uint64_t t1 = get_cycles();
        int res1 = calculateFunction(a1, 'X', size);
        uint64_t t2 = get_cycles();
        int res2 = sseFunction(a1, 'X', size);
        uint64_t t3 = get_cycles();
        
        printf("%d %d %d ", (int)(t2 - t1), (int)(t3 - t2), res1);

        if(res1 != res2) {
            printf("INCORRECT\n");
        }
        else {
            printf("CORRECT\n");
        }
    }
    return 0;
}
